import urllib.error
import re

import feedparser

from xml.sax import SAXException

from series.logging import Logging
from golgi.config import configurable
from tek.tools import find

from series.get.handler import ReleaseHandler
from series.get.model.feed_entry import FeedEntryFactory
from series.get.model.feed_entry import FeedEntryHasher
from series.get.model.release import ReleaseFactory

feedparser.PREFERRED_XML_PARSERS = []


@configurable(series=['monitor'], get=['rss_urls', 'rss_interval'])
class FeedPoller(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, self._rss_interval,
                                         'feed poller')
        self._rss = None
        self._etags = [None] * len(self._rss_urls)
        self._modified = [None] * len(self._rss_urls)
        self._release_fact = ReleaseFactory()
        self._hashes = set()
        self._feed_entries = []

    def reconstruct_entries(func):
        def wrapper(dl, *a, **kw):
            ret = func(dl, *a, **kw)
            dl._reconstruct_entries()
            return ret
        return wrapper

    @reconstruct_entries
    def load_from_xml_file(self, fname):
        self._rss = feedparser.parse(fname)

    @reconstruct_entries
    def load_from_url(self, url, index):
        try:
            self._rss = feedparser.parse(url, etag=self._etags[index],
                                         modified=self._modified[index])
            self._etags[index] = self._rss.get('etag', None)
            self._modified[index] = self._rss.get('modified_parsed', None)
        except (urllib.error.URLError, urllib.error.HTTPError,
                SAXException) as e:
            self.log.debug(e)

    def _check(self):
        for index, url in enumerate(self._rss_urls):
            self.load_from_url(url, index)

    def _reconstruct_entries(self):
        if self._rss and self._rss.entries:
            hasher = FeedEntryHasher(self._rss.feed.link)
            valid = lambda e: (bool(e) and not hasher(e) in self._hashes)
            valid_entries = list(filter(valid, self._rss.entries))
            self._hashes |= set([hasher(e) for e in valid_entries])
            fact = FeedEntryFactory(self._rss.feed.link)
            self._feed_entries = fact.process_items(reversed(valid_entries))
            text = 'Found {} series entries in the feed "{}".'
            self.log.info(text.format(len(self._feed_entries), fact.domain))
            self._update_monitors()

    def _update_monitors(self):
        for entry in filter(self._is_monitored, self._feed_entries):
            existing = self._find_release(entry)
            if existing is None:
                self._create_monitor(entry)
            else:
                self._releases.add_links_from_feed_entry(existing, entry)

    def _create_monitor(self, entry):
        monitor = self._release_fact.monitor_from_entry(entry)
        old_entry = self._find_nuke(entry)
        if old_entry is not None:
            if not entry.is_fix:
                self._update(monitor, downloaded=old_entry.downloaded)
                self._update(monitor, archived=old_entry.archived)
            old_entry.nuke(monitor)
        self._releases.add(monitor)
        self.log.info('Added monitor for {!s}'.format(monitor.release))

    def _find_release(self, entry):
        return find(lambda m: m.release == entry.release, self._releases.all)

    def _find_nuke(self, entry):
        return find(lambda m: m.release.is_same_episode(entry.release),
                    self._releases.all)

    def _is_monitored(self, entry):
        name = entry.release.name
        if re.search(r'_\d{4}$', name):
            name = name.rsplit('_', 1)[0]
        return name in self._monitor

__all__ = ['FeedPoller']



''' configuration for library '''

from golgi.config import (PathListConfigOption, PathConfigOption,
                        ListConfigOption, FloatConfigOption)
from golgi.config.options import DictConfigOption

metadata = dict(parents=['series.etvdb'])


def reset_config():
    ''' Initialize the config. '''
    season_dir = '{name}/s{season}/'
    episode = '{name}_{season:0>2}x{episode:0>2}.{ext}'
    path_template = season_dir + episode
    subtitle_path_template = season_dir + 'sub/' + episode
    movie_path_template = '{title}.{ext}'
    movie_subtitle_path_template = 'sub/{title}.{ext}'
    run_help = ('Comma-separated list of components to start (rest_api, ' +
                'metadata)')
    omit_help = 'List of components not to start'
    resume_before_help = (
        'Relative value in [0, 1], when an episode has been stopped at a '
        'position before this progress, it is resumed there when starting '
        'again')
    return {
        'library': dict(
            collection_paths=PathListConfigOption(),
            movie_collection_paths=PathListConfigOption(),
            path_template=path_template,
            subtitle_path_template=subtitle_path_template,
            movie_path_template=movie_path_template,
            movie_subtitle_path_template=movie_subtitle_path_template,
            rest_api_port=8111,
            rest_api_host='0.0.0.0',
            db_path=PathConfigOption(),
            name_formatter='',
            run=ListConfigOption([], help=run_help),
            omit=ListConfigOption([], help=omit_help),
            metadata_interval=300,
        ),
        'player': dict(
            display=None,
            path='mplayer2',
            args=DictConfigOption(dict(fs='yes')),
            resume_before=FloatConfigOption(0, help=resume_before_help),
            subtitles=True,
            mplayer_extra_args=DictConfigOption({}),
            mpv_extra_args=DictConfigOption(dict(input_default_bindings='yes',
                                                 input_vo_keyboard='yes')),
            player_type='mplayer',
        ),
        'library_client': dict(
            rest_api_url='http://localhost', rest_api_port=8111
        )
    }

__all__ = ['reset_config', 'metadata']



from datetime import datetime

from series.get.handler import ShowHandler
from series.get.tvdb import Tvdb

from golgi.config import configurable
from series.logging import Logging


@configurable(series=['monitor'], show_planner=['check_interval', 'show_db'])
class ShowPlanner(ShowHandler, Tvdb):

    def __init__(self, releases, shows, *a, **kw):
        super().__init__(shows, 5, 'show planner', *a,
                                          **kw)
        self._releases = releases
        self._shows_initialized = False

    def _sanity_check(self):
        if not self._shows_initialized:
            self._init()

    def _init(self):
        for name in self._monitor:
            self._init_show(name)
        self._shows_initialized = True

    def _init_show(self, name):
        if not self._shows.name_exists(name):
            self._add_show(name)

    def _add_show(self, name):
        self.log.debug('Adding show for "{}"'.format(name))
        show = self.tvdb.show(name)
        if show:
            self._shows.add(name, show)
        else:
            self.log.debug('Adding show failed.')

    def _handle(self, show):
        show.last_check = datetime.now()
        rage = self.tvdb.show(show.name, show.tvdb_id)
        if rage is None:
            self.log.error(
                'Show couldn\'t be found anymore: {}'.format(show.name)
            )
        else:
            self._update_show(show, rage)

    def _update_show(self, show, rage):
        self.log.debug('Updating show {}'.format(rage.name))
        data = {}
        if rage.ended:
            data['ended'] = True
            self.log.warn('Show has ended: {}'.format(show.name))
        else:
            data = self._fetch_next_episode(show, rage)
        if show.tvdb_id is None:
            data.update(**self.tvdb.show_id_update_param(show))
        self._shows.update(show.id, data)

    def _fetch_next_episode(self, show, rage):
        data = {}
        self.log.debug('Fetching next episode for "{}"'.format(show.name))
        latest = rage.latest_episode
        if latest is not None:
            data.update(latest_episode=latest.number,
                        latest_season=latest.season)
        date = self.tvdb.next_episode_date(rage)
        if date is not None:
            data['next_episode_date'] = date
            enum = self.tvdb.next_episode_enum(rage)
            data['season'], data['next_episode'] = enum
            self.log.debug('Found season {} episode {}'.format(*enum))
        else:
            self.log.debug('Fetching next episode failed.')
        return data

    def _qualify(self, show):
        return (not show.has_next_episode and not show.ended
                and show.can_recheck(self._check_interval))

__all__ = ['ShowPlanner']



import threading
import time
import importlib

from golgi.run import SignalManager

from tryp import List, Map
from tryp.util.string import camelcaseify
from series.logging import Logging


class App(threading.Thread, Logging):
    _components = []  # type: List[Handler]

    def __init__(self, c_module, run, omit, c_args=(), interval=0.2,
                 name='series'):
        super().__init__(name=name)
        self._c_module = c_module
        self._c_args = c_args
        self._interval = interval
        self.name = name
        self._running = False
        self.components = List()
        self.component_map = Map()
        self._init_components(run, omit)

    def run(self):
        self.register_sigint_handler()
        self.pre()
        self.main_loop()

    def pre(self):
        self.start_components()
        self.prepare()

    def prepare(self):
        pass

    def start_components(self):
        for component in self.components:
            component.start()

    def main_loop(self):
        self._running = True
        self.log.info('{} running.'.format(self.name))
        while self._running:
            time.sleep(self._interval)

    def _init_components(self, run, omit):
        components = self._components
        if run:
            components = [c for c in components if c in run]
        if omit:
            components = [c for c in components if c not in omit]
        for component in components:
            self._setup_component(component)

    def _setup_component(self, component):
        def error(exc):
            msg = 'Invalid component name: {} ({})'.format(component, exc)
            self.log.error(msg)
        mod_name = 'series.{}.{}'.format(self._c_module, component)
        try:
            module = importlib.import_module(mod_name)
        except ImportError as e:
            error(e)
        else:
            try:
                Component = getattr(module, camelcaseify(component))
            except AttributeError as e:
                error(e)
            else:
                instance = Component(*self._c_args)
                setattr(self, component, instance)
                self.components.append(instance)
                self.component_map[component] = instance

    def register_sigint_handler(self):
        SignalManager.instance.sigint(self.interrupt)

    def interrupt(self, signum=10, frame=None):
        self._stop_components()
        self._join_threads()
        self._cleanup()
        self._running = False
        self.log.info('{} shut down due to SIGINT.'.format(self.name))

    def _stop_components(self):
        for component in self.components:
            component.stop()

    def _join_threads(self):
        for component in self.components:
            if component.is_alive():
                component.join()

    def _cleanup(self):
        pass

__all__ = ['App']



import re

from tek.errors import ParseError


def domain(url):
    """ Extract the domain name out of url.
    """
    if isinstance(url, str):
        domain_match = re.match(r'https?://(?:www\.)?([^/]+)\.[^/]+', url)
        return domain_match.group(1) if domain_match else ''
    else: raise ParseError('Invalid input for domain(): {}'.format(url))


__all__ = ['domain']



import os

from golgi.config import configurable
from tek.tools import wait_for

from series.library.playback_observer import PlaybackObserver


@configurable(player=['display', 'path', 'subtitles', 'args', 'resume_before',
                      'mplayer_extra_args', 'mpv_extra_args', 'player_type'])
class Player(object):

    def __init__(self, library):
        self._library = library
        if self._display:
            os.environ['DISPLAY'] = self._display
        self._mplayer = None
        self._target = None
        self._observer = None
        self._extra_args = []
        self._setup_player_type()

    def _setup_player_type(self):
        if self._player_type == 'mplayer':
            try:
                from series.library.player.mplayer import MPlayer
                from mplayer import Player
            except ImportError:
                pass
            else:
                self._Player = MPlayer
                Player.exec_path = self._path
                self._extra_args = self._mplayer_extra_args
        elif self._player_type == 'mpv':
            try:
                from series.library.player.mpv import MPV
            except ImportError:
                pass
            else:
                self._Player = MPV
                self._extra_args = self._mpv_extra_args

    @property
    def target(self):
        return self._target

    @target.setter
    def target(self, epi):
        self.stop()
        self._target = epi

    def start(self):
        self._check_file()
        self._mplayer = self._Player(self._all_args)
        self._observer = PlaybackObserver(self._mplayer, self._library,
                                          self._target, self)
        self._mplayer.loadfile(str(self._file))
        self._observer.start()
        self._resume()

    def _resume(self):
        wait_for(lambda: self.running)
        pos = self._target.resume_position
        length = self._wait_for_prop('length')
        if pos and length and pos < self._resume_before * length:
            self._set_pos(pos)

    def _check_file(self):
        if not (self._file and self._file.exists()):
            raise OSError('File not found: {}'.format(self._file))

    def toggle_pause(self):
        if self.running:
            self._mplayer.pause()

    def seek(self, value):
        if self.running:
            elapsed = self._wait_for_prop('time_pos')
            if elapsed is not None:
                new_value = elapsed + value
                self._set_pos(new_value)

    def seek_to_ratio(self, value):
        if self.running:
            new_value = value * self.length
            self._set_pos(new_value)

    def _set_pos(self, new_value):
        new_value = max(0, new_value)
        new_value = min(self.length, new_value)
        self._mplayer.time_pos = new_value

    def sub_delay(self, value):
        if self.running:
            self._mplayer.sub_delay += value

    def change_volume(self, value):
        if self.running:
            current = self._player_prop('volume')
            self._mplayer.volume = min(100, max(0, current + value))

    def stop(self):
        if self._observer is not None:
            self._observer.stop()
            self._observer.join()
            self._observer = None
        if self._mplayer is not None:
            self._mplayer.stop()
            self._mplayer = None

    def toggle_info(self):
        if self.running:
            current = self._mplayer.osd_level()
            self._mplayer.osd(0 if current > 1 else 3)

    @property
    def elapsed(self):
        return self._player_prop('time_pos')

    @property
    def length(self):
        return self._player_prop('length')

    @property
    def remaining(self):
        return self.length - self.remaining

    @property
    def _file(self):
        return self._library.video_path(self.target)

    @property
    def running(self):
        return self._mplayer is not None and self._mplayer.running

    @property
    def _all_args(self):
        args = self._args.copy()
        args.update(self._subtitle_args)
        args.update(self._subfps_args)
        args.update(self._extra_args)
        return args

    @property
    def _subtitle_args(self):
        return (dict(sub=self._library.subtitle_path(self._target))
                if self._subtitles and self._target
                else [])

    @property
    def _subfps_args(self):
        value = self._target.subfps_fallback
        return dict(subfps=value) if value else []

    def _player_prop(self, name, default=0):
        return getattr(self._mplayer, name, default)

    def _wait_for_prop(self, name):
        value = None

        def poll():
            nonlocal value
            value = self._player_prop(name, None)
            return value
        wait_for(poll, 5, 0.1)
        return value

    def same_episode(self, observer):
        return observer is self._observer

__all__ = ['Player']



from golgi.config import ListConfigOption

metadata = dict(parents=['series'])


def reset_config():
    return {
        'store_episode': dict(
            episodes=ListConfigOption(),
            season_regex='s(\d+)',
            pretend=False,
            remove_prefixes=ListConfigOption(),
            series_name=None,
            season_name=None,
            match_threshold=0.7,
            overwrite=False,
            ask_series=True,
        )
    }



from series.get.handler import ShowHandler
from series.get.tvdb import Tvdb

from series.logging import Logging


class ShowScheduler(ShowHandler, Tvdb):

    def __init__(self, releases, shows, **kw):
        super().__init__(shows, 5, 'show scheduler',
                                            cooldown=3600, **kw)
        self._releases = releases

    def _handle(self, show):
        latest = self._latest(show)
        if latest is None:
            self._schedule(show, show.season, show.next_episode)
        else:
            for episode in range(latest.episode + 1, show.current_episode + 1):
                self._schedule(show, show.current_season, episode)

    def _qualify(self, show):
        latest = self._latest(show)
        if latest is None:
            return show.next_episode_imminent
        else:
            return show.current_episode > latest.episode

    def _latest(self, show):
        return self._releases.latest_for_season(show.canonical_name,
                                                show.current_season)

    def _schedule(self, show, season, episode):
        airdate = self.tvdb.airdate(show, season, episode)
        msg = 'Scheduling release "{} {}x{}" on {}'.format(
            show.name,
            season,
            episode,
            airdate
        )
        self.log.info(msg)
        self._releases.create(show.canonical_name, season, episode, airdate)
        self._commit()

__all__ = ['ShowScheduler']



import os
import re
import mimetypes
from pathlib import Path

from golgi import ConfigClient
from golgi.config import configurable

from tryp import LazyList

from series.errors import MissingMetadata


def dir_info(dir):
    rex = ConfigClient('series')('series_dir') + '(/([^/]+))?(/s(\d+))?(/.*)?'
    match = re.match(rex, dir)
    if match:
        return match.group(2), match.group(4)
    else:
        return None, None

styles = (r's?{}e{}', r'{}x{}', r's?{}.*e{}', r'(\d\d)(\d\d){}?{}?',
          r'(\d)(\d\d)', r'(){}{}?', r'{}.*?{}')
digits = (r'(\d\d?)', r'(\d{1,3})')
regexes = [s.format(*digits) for s in styles]


def episode_enumeration_match(filename):
    custom_rex = ConfigClient('series')('enumeration_regex')
    if custom_rex:
        return re.search(custom_rex, str(filename))
    else:
        if isinstance(filename, Path):
            filename = filename.name
        searcher = lambda s: re.search(s, filename, re.I)
        rex = LazyList(map(searcher, regexes))
        return rex.find(lambda a: a) | None


def episode_enumeration(filename):
    custom_rex = ConfigClient('series')('enumeration_regex')
    if isinstance(filename, str):
        match = episode_enumeration_match(filename)
        if match:
            if custom_rex:
                groups = match.group('season'), match.group('episode')
            else:
                groups = match.groups()[:2]
            if all(s.isdigit() for s in groups):
                return list(map(int, groups))
    return None, None


def episode_number(filename):
    return episode_enumeration(filename)[1]


def season_dir(series, season):
    return ConfigClient('series')('series_dir') / series / 's{}'.format(season)


def get_release(series, season, number):
    filename = '.release_{}'.format(str(number).zfill(2))
    release_file = season_dir(series, season) / filename
    if not release_file.is_file():
        return None
    else:
        with release_file.open() as f:
            return f.readlines()[0].rstrip()


def latest_season_dir(series):
    series_dir = os.path.join(ConfigClient('series')('series_dir'), series)
    is_series = lambda dir: re.match('s\d\d?', dir)
    latest = sorted(filter(is_series, os.listdir(series_dir)))[-1]
    return os.path.join(series_dir, latest)


def make_series_name(string):
    string = string.strip('-[_. ]+')
    parts = [_f for _f in re.split(r'[_. ]+', string) if _f]
    string = '_'.join(parts)
    return string.lower()


def episode_metadata(filename):
    filename = Path(filename)
    result = episode_enumeration_match(filename)
    if result:
        groups = list(result.groups())
        index = result.start()
        enum_rex = ConfigClient('series')('enumeration_regex')
        if enum_rex:
            series_name = result.group('name')
            season, number = result.group('season'), result.group('episode')
        else:
            series_name = make_series_name(str(filename.name)[:index])
            season, number = groups[:2]
        return series_name, season, number


def is_video(_file):
    guess = mimetypes.guess_type(_file)
    return (bool(guess) and isinstance(guess[0], str) and
            guess[0].split('/')[0] == 'video')


def is_episode(_file):
    return isinstance(episode_enumeration(_file)[0], int)


def canonicalize(name):
    return name.lower().replace(' ', '_')


def is_date_enum(enum):
    return bool(re.match('\d{4}\.\d{1,2}\.\d{1,2}$', enum))


def convert_date_enum(name, enum):
    year, month, day = enum.split('.')
    return '{}_{}'.format(name, year), 'x'.join((month, day))


class EpisodeMetadata(list):

    def __init__(self, series, season, episode, release=None, extension=None):
        super().__init__([series, int(season),
                                               int(episode)])
        self.release = release
        self.extension = extension

    @property
    def series(self):
        return self[0]

    @property
    def season(self):
        return self[1]

    @property
    def season_zfill(self):
        return str(self.season).zfill(2)

    @property
    def episode(self):
        return self[2]

    @property
    def episode_zfill(self):
        return str(self.episode).zfill(2)

    @property
    def all(self):
        return self + [self.release]

    @property
    def filename(self):
        file = '{0}_{1}x{2}.{3}'
        if self.extension is None:
            raise MissingMetadata('No extension in EpisodeMetadata!')
        return file.format(self.series, self.season_zfill, self.episode_zfill,
                           self.extension)

    @property
    def local_path(self):
        path = os.path.join(self.series, 's{}'.format(self.season))
        if self.extension == 'srt':
            path = os.path.join(path, 'sub')
        return os.path.join(path, self.filename)

    @property
    def is_subtitle(self):
        return self.extension == 'srt'


class SubtitleMetadata(EpisodeMetadata):

    def __init__(self, series, season, episode, release=None, extension='srt'):
        super().__init__(series, season, episode,
                                               release=release,
                                               extension=extension)

    @property
    def is_subtitle(self):
        return True


@configurable(series=['series_dir', 'enumeration_regex'])
class EpisodeMetadataFactory(object):

    def from_filename(self, path):
        series_name, season, number = episode_metadata(path)
        series_name = canonicalize(series_name)
        release = get_release(series_name, season, number)
        extension = path.name.rsplit('.', 1)[-1]
        typ = SubtitleMetadata if extension == 'srt' else EpisodeMetadata
        return typ(series_name, season, number, release, extension)


__all__ = ['EpisodeMetadataFactory', 'dir_info', 'episode_enumeration_match',
           'episode_enumeration', 'episode_number', 'season_dir',
           'get_release', 'latest_season_dir', 'make_series_name',
           'episode_metadata', 'is_video', 'is_episode', 'canonicalize',
           'EpisodeMetadata', 'SubtitleMetadata']



import re

from golgi.config import configurable

from series.logging import Logging
from tek.tools import find
from golgi.io.terminal import terminal as term, ColorString

from tryp import Map, List, __, F

from series.api_client import ApiClient as Base
from series.get import format_explain


def is_error(response):
    return isinstance(response, dict) and 'error' in response


@configurable(get_client=['rest_api_port', 'rest_api_url'])
class ApiClient(Base):
    command = Base.command

    def _find_release_id(self, series, season, episode, url=None):
        if episode is not None:
            data = dict(series=series, season=season, episode=episode)
            _id = self.get('release_id', body=data)
        elif isinstance(series, int) or (isinstance(series, str) and
                                         series.isdigit()):
            _id = int(series)
        elif url is not None:
            dls = self.get('download/pending')
            release = find(lambda r: re.search(r[1], url, re.I), dls)
            _id = release[0] if release else -1
        else:
            _id = -1
        return _id

    def _with_release(self, series, season, episode, f, url=None):
        _id = self._find_release_id(series, season, episode, url)
        if _id != -1:
            response = f(_id)
            if not is_error(response):
                self.log.info(response)
            else:
                _id = -1
                self.log.error('Command failed')
        else:
            self.log.error('No matching release found')
        return _id

    def _with_release_args(self, f, fr, *args, **kw):
        series, season, episode, rest = self._parse_episode_args(args)
        return self._with_release(series, season, episode, F(f, rest), **kw)

    def _release_put(self, series, season, episode, path=None, url=None,
                     **args):
        sub = '/{}'.format(path) if path else ''
        go = lambda _id: self.put('release/{}{}'.format(_id, sub), **args)
        return self._with_release(series, season, episode, go, url)

    @command('url (id || series season episode)', 'Add the url to the ' +
             'specified release\'s links')
    def add_link(self, url, series=None, season=None, episode=None):
        return self._release_put(series, season, episode, 'link', url=url,
                                 body=dict(url=url))

    @command('count', 'Describe the state of the latest `count` releases,'
             ' default 5')
    def explain_old(self, count=5):
        output = self.get('release/explain_old/{}'.format(count))
        for line in output:
            self.log.info(line)
        return output

    @command('id', 'Set the \'downloaded\' flag for the specified release')
    def mark_downloaded(self, _id):
        return self._downloaded_flag(_id, True)

    @command('id', 'Unset the \'downloaded\' flag for the specified release')
    def mark_not_downloaded(self, _id):
        return self._downloaded_flag(_id, False)

    def _downloaded_flag(self, _id, value):
        success = self.put('release/{}'.format(_id),
                           body=dict(downloaded=value, archived=value))
        if success:
            self.log.info('Success!')
        else:
            self.log.error('Release not found!')
        return bool(success)

    @command('[regex]', 'Display id, name, season and episode for all ' +
             'releases matching the regex (default all)')
    def list(self, regex=''):
        matches = self.get('release', body=dict(regex=regex))
        if matches:
            text = 'id #{}: {}'
            for id, description in matches:
                self.log.info(text.format(id, description))
        else:
            self.log.info('No matching release found.')
        return matches

    @command('series season episode', 'Create an empty release with the ' +
             'supplied metadata')
    def create_release(self, series, season, episode):
        response = self.post('release/{}/{}/{}'.format(series, season,
                                                       episode))
        self.log.info(response)
        return response

    @command('series season episode', 'Delete the release matching the ' +
             'supplied metadata')
    def delete_release(self, series, season, episode):
        response = self.delete('release/{}/{}/{}'.format(series, season,
                                                         episode))
        self.log.info(response)
        return response

    @command('series season', 'Add the specified season of series to the db')
    def add_season(self, name, season):
        response = self.post('season', body=dict(name=name, season=season))
        self.log.info(response)
        return response

    @command('name', 'Add the specified show')
    def add_show(self, name):
        response = self.post('show', body=dict(name=name))
        self.log.info(response)
        return response

    @command('canonical_name|id', 'Delete the specified show')
    def delete_show(self, name):
        response = self.delete('show', body=dict(name=name))
        self.log.info(response)
        return response

    @command('[regex]', 'List show names matching the regex')
    def list_shows(self, regex=''):
        matches = self.get('show', body=dict(regex=regex))
        if matches:
            text = 'id #{}: {}'
            for id, description in matches:
                self.log.info(text.format(id, description))
        else:
            self.log.info('No matching show found.')
        return matches

    def print_shows(self, shows):
        colors = {
            0: term.blue,
            1: term.green,
            2: term.yellow,
            3: term.red,
        }
        if shows:
            for name, nepi, rel, status in shows:
                term.push([ColorString('>> ', term.red),
                           ColorString(name, term.bold)])
                term.push([ColorString(' | ', term.green), nepi])
                if rel:
                    col = colors.get(status, term.black)
                    term.push([ColorString(' | ', term.green),
                               ColorString(rel, col)])
                self.log.info('')
        else:
            self.log.info('No matching show found.')
        return shows

    @command('[regex]', 'Extended info for shows matching regex')
    def shows(self, regex=''):
        return self.print_shows(self.get('show/info', body=dict(regex=regex)))

    @command('[regex]', 'List upcoming releases for shows matching regex')
    def next(self, regex=''):
        return self.print_shows(self.get('show/next', body=dict(regex=regex)))

    @command('[regex]', 'List current releases for shows matching regex')
    def ready(self, regex=''):
        return self.print_shows(self.get('show/ready', body=dict(regex=regex)))

    @command('[regex]', 'List downloaded releases for shows matching regex')
    def done(self, regex=''):
        return self.print_shows(self.get('show/done', body=dict(regex=regex)))

    @command('id || series season episode', 'Reset a release and mark its ' +
             'torrent as dead, forcing download of a different torrent')
    def reset_torrent(self, series=None, season=None, episode=None):
        return self._release_put(series, season, episode, 'reset_torrent')

    @command('id || series season episode', 'Print release info')
    def show_release(self, series=None, season=None, episode=None):
        go = lambda _id: self.get('release/{}'.format(_id))
        return self._with_release(series, season, episode, go)

    def _parse_episode_args(self, args):
        a = List.wrap(args)
        if a.lift(2).exists(__.isdigit()):
            series, season, episode = a[:3]
            rest = a[3:]
        else:
            series, season, episode = a[0], None, None
            rest = a[1:]
        return series, season, episode, rest

    @command('(id || series season episode) [key=value ...]',
             'Update release data')
    def update_release(self, *args):
        series, season, episode, rest = self._parse_episode_args(args)
        data = Map(rest.map(__.split('=')))
        return self._release_put(series, season, episode, '', body=data)

    @command('id || series season episode', 'Reset cooldown time for release')
    def activate_release(self, *args):
        series, season, episode, _ = self._parse_episode_args(args)
        return self._release_put(series, season, episode, 'activate')

    @command('[id || series season episode] [services]', 'Explain why the ' +
             'csv-specified services treat the release as they do')
    def explain(self, *args):
        def rest(a):
            return a.head.map(lambda b: dict(body=dict(services=b))) | dict()
        def go(r, id):
            resp = self.get('release/{}/explain'.format(id), **rest(r))
            return resp if is_error(resp) else format_explain(resp)
        self._with_release_args(go, rest, *args)

__all__ = ['ApiClient']



import tryp.logging
from tryp.logging import tryp_logger
from tryp.lazy import lazy

log = series_root_logger = tryp_logger('series')


def series_logger(name: str):
    return series_root_logger.getChild(name)


class Logging(tryp.logging.Logging):

    @lazy
    def _log(self) -> tryp.logging.Logger:  # type: ignore
        return series_logger(self.__class__.__name__)

__all__ = ('series_logger', 'Logging')



# -*- coding: utf-8 -*-

import re
import os
import subprocess
import shutil
import shlex
from pathlib import Path

from series.logging import Logging
from golgi.config import configurable

from tryp import _, List

from series import store_episode
from series.store_episode import EpisodeHandler as EpisodeStore
from series.get.errors import ArchiverError
from series.get.handler import ReleaseHandler, R


@configurable(series=['series_dir'],
              get=['path_template', 'archive_exec', 'archive_exec_args'])
class Archiver(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 5, 'archiver')

    @property
    def _conditions(self):
        return ((~(R('unknown') | R('archived') | R('archiver_error'))) &
                R('downloaded'))

    def _handle(self, monitor):
        release = monitor.release
        self.log.info('Archiving {!s}.'.format(release))
        try:
            if not monitor.download_path:
                raise ArchiverError('Can\'t archive: no download path!')
            self._archive(monitor)
        except store_episode.errors.UnknownSeries as e:
            self.log.error(e)
            self._update(monitor, unknown=True)
        except ArchiverError as e:
            self.log.error(e)
            self._update(monitor, archiver_error=True)
        else:
            self._releases.mark_episode_archived(monitor)
        self._commit()

    def _archive(self, monitor):
        if shutil.which(self._archive_exec):
            self._external(monitor)
        else:
            dest_path = self._setup_paths(monitor)
            self._store(monitor, dest_path)

    def _external(self, monitor):
        cmdline = [self._archive_exec] + self._archiver_params(monitor)
        proc = subprocess.Popen(cmdline)
        try:
            retval = proc.wait(timeout=20)
        except subprocess.TimeoutExpired:
            raise ArchiverError('External archiver timed out!')
        else:
            if retval != 0:
                raise ArchiverError('External archiver failed!')

    def _archiver_params(self, monitor):
        arg_str = self._archive_exec_args.format(
            series_dir=self._series_dir,
            download_path=shlex.quote(monitor.download_path),
            name=re.escape(monitor.release.name),
            season=monitor.release.season,
            episode=monitor.release.episode,
        )
        try:
            return shlex.split(arg_str)
        except ValueError as e:
            self.log.error('formatting archiver params: {}'.format(e))
            self.log.error('release {}'.format(str(monitor.release)))
            self.log.error('format string {}'.format(arg_str))
            raise ArchiverError('Couldn\'t assemble archiver params')

    def _store(self, monitor, dest_path):
        handler = EpisodeStore(ask=False, series_name=monitor.release.name)
        handler.add_job(Path(monitor.download_path), dest_path)
        if not handler.store():
            raise ArchiverError('Built-in archiver failed!')

    def _setup_paths(self, monitor):
        local_path = self._format_path(monitor)
        dest_path = self._series_dir / local_path
        target_dir = dest_path.parent
        self._check_target_dir(target_dir)
        self._check_downloaded_file(monitor, dest_path)
        return dest_path

    def _check_downloaded_file(self, monitor, dest_path):
        if not os.path.isfile(monitor.download_path):
            text = 'Downloaded release "{!s}" missing!'
            if dest_path.is_file():
                text += '\nHowever, it seems to be archived.'
                self._update(monitor, archived=True)
            raise ArchiverError(text.format(monitor.release))

    def _check_target_dir(self, target_dir):
        if not target_dir.exists():
            target_dir.mkdir(parents=True)
        elif not target_dir.is_dir():
            raise ArchiverError('Not a directory: {}'.format(target_dir))

    def _format_path(self, monitor):
        parts = monitor.download_path.rsplit('.', 1)
        release = monitor.release
        if len(parts) <= 1:
            self.log.error('Invalid filename. Assuming \'mkv\' extension.')
            extension = 'mkv'
        else:
            extension = parts[1]
        parameters = dict(
            name=release.name,
            season=release.season,
            episode=release.episode,
            ext=extension,
        )
        local = self._path_template.format(**parameters)
        return self._series_dir / local

__all__ = ['Archiver']
